/*
 * Copyright (C) 2003, 2004, 2006, 2008 Robert Lougher <rob@jamvm.org.uk>.
 *
 * This file is part of JamVM.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

package jamvm.java.lang;

import java.util.jar.JarFile;
import java.util.jar.Manifest;
import java.util.jar.Attributes.Name;
import java.lang.reflect.Method;

public class JarLauncher {
    public static void main(String[] args) {
        String jarFile = System.getProperty("java.class.path", null);

        try {
            JarFile jar = new JarFile(jarFile);
            Manifest manifest = jar.getManifest();

            if(manifest == null)
                System.err.println("Couldn't read manifest from " + jar.getName() + " (invalid jar file?).");
            else {
                String mainAttr = manifest.getMainAttributes().getValue(Name.MAIN_CLASS);

                if(mainAttr == null)
                    System.err.println("Couldn't find Main-Class attribute in " +
                                        jar.getName() + " Manifest.");
                else {
                    ClassLoader sysLoader = ClassLoader.getSystemClassLoader();
                    Class mainClass = Class.forName(mainAttr, true, sysLoader);
                    Method meth = mainClass.getMethod("main", new Class[] {args.getClass()});

                    meth.setAccessible(true);
                    meth.invoke(null, new Object[] {args});
                }
            }
        } catch(Exception e) {
            e.printStackTrace();
        }
    }
}
